// cache.h
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
// Author: allauzen@cs.nyu.edu (Cyril Allauzen)
//
// \file
// Cache representation.

#ifndef KERNEL_LIB_CACHE_H__
#define KERNEL_LIB_CACHE_H__

#include "kernel/lib/kernel.h"

namespace kernel {

typedef unsigned int uint32;

const uint32 kCacheValue = 0x01;
const uint32 kCacheFeature = 0x02;
const uint32 kCacheNorm = 0x04;

template <class X, class F>
class Cache {
 public:
  virtual ~Cache() {}
  virtual uint32 CacheType() const = 0;
  virtual bool HasValue(size_t i, size_t j) const = 0;
  virtual bool HasFeature(size_t i) const = 0;
  virtual bool HasNorm(size_t i) const = 0;
  virtual double GetValue(size_t i, size_t j) const = 0;
  virtual F* GetFeature(size_t i) const = 0;
  virtual double GetNorm(size_t i) const = 0;
  virtual void SetValue(size_t i, size_t j, double x) = 0;
  virtual void SetFeature(size_t i, F* f) = 0;
  virtual void SetNorm(size_t i, double x) = 0;
  virtual void Clear() = 0;
}; //use const pointers for features


template <class X, class F>
class BasicCache : public Cache<X,F> {
 public:
  typedef X Object;
  typedef F Feature;

  BasicCache(uint32 cache_type = kCacheFeature) : type_(cache_type) {}

  virtual ~BasicCache() {
    for (size_t i = 0; i < features_.size(); ++i)
      if (features_[i]) delete features_[i];
  }

  uint32 CacheType() const {
    return type_;
  }

  bool HasValue(size_t i, size_t j) const {
    size_t x = index(i,j);
    if (x < has_value_.size())
      return has_value_[x];
    return false;
  }

  double GetValue(size_t i, size_t j) const {
    size_t x = index(i,j);
    if ((has_value_.size() <= x) || (has_value_[x] == false))
      LOG(FATAL) << "BasicCache: no value cached";
    return values_[x];
  }

  void SetValue(size_t i, size_t j, double v) {
    size_t x = index(i,j);
    while (has_value_.size() <= x) has_value_.push_back(false);
    has_value_[x] = true;
    while (values_.size() <= x) values_.push_back(0.0F);
    values_[x] = v;
  }

  bool HasFeature(size_t i) const {
    if (i < has_feature_.size())
      return has_feature_[i];
    return false;
  }

  F* GetFeature(size_t i) const {
    if ((has_feature_.size() <= i) || (has_feature_[i] == false))
      LOG(FATAL) << "BasicCache: no features cached";
    return features_[i];
  }

  void SetFeature(size_t i, F *f) {
    while (has_feature_.size() <= i) has_feature_.push_back(false);
    has_feature_[i] = true;
    while (features_.size() <= i) features_.push_back(0);
    if (features_[i]) delete features_[i];
    features_[i] = f;
  }

  bool HasNorm(size_t i) const {
    if (i < has_norm_.size())
      return has_norm_[i];
    return false;
  }

  double GetNorm(size_t i) const {
    if ((has_norm_.size() <= i) || (has_norm_[i] == false))
      LOG(FATAL) << "BasicCache: no norm cached";
    return norms_[i];
  }

  void SetNorm(size_t i, double v) {
    while (has_norm_.size() <= i) has_norm_.push_back(false);
    has_norm_[i] = true;
    while (norms_.size() <= i) norms_.push_back(0.0F);
    norms_[i] = v;
  }

  static BasicCache<X,F> *Read(istream &strm, const KernelReadOptions &opts) {
    uint32 type;
    ReadType(strm, &type);
    if (!strm) return 0;

    BasicCache<X,F> *cache = new BasicCache<X,F>(type);

    bool read_cache;
    ReadType(strm, &read_cache);
    if (strm && read_cache) {
      if (type & kCacheValue) {
        int64 nhv;
        ReadType(strm, &nhv);
        cache->has_value_.reserve(nhv);
        while (cache->has_value_.size() < nhv) {
          bool b;
          ReadType(strm, &b);
          cache->has_value_.push_back(b);
        }
        int64 nv;
        ReadType(strm, &nv);
        cache->values_.reserve(nv);
        while (cache->values_.size() < nv) {
          double x;
          ReadType(strm, &x);
          cache->values_.push_back(x);
        }
      }
    }

    if (!strm) {
      delete cache;
      return 0;
    }
    return cache;
  }

  bool Write(ostream &strm, const KernelWriteOptions &opts) const {
    WriteType(strm, type_);

    WriteType(strm, opts.write_cache);
    if (opts.write_cache) {
      if (type_ & kCacheValue) {
        int64 nhv = has_value_.size();
        WriteType(strm, nhv);
        for (int64 i = 0; i < nhv; i++)
          WriteType(strm, has_value_[i]);
        int64 nv = values_.size();
        WriteType(strm, nv);
        for (int64 i = 0; i < nv; i++)
          WriteType(strm, values_[i]);
      }
    }
    return strm;
  }

  void Clear() {
    features_.clear();
    has_feature_.clear();
    values_.clear();
    has_value_.clear();
    norms_.clear();
    has_norm_.clear();
  }

 private:
  size_t index(size_t i, size_t j) const {
    if ( i <= j)
      return (j * (j + 1))/2 + i;
    else
      return (i * (i + 1))/2 + j;
  }

  uint32 type_;
  vector<Feature*> features_;
  vector<bool> has_feature_;
  vector<double> values_;
  vector<bool> has_value_;
  vector<double> norms_;
  vector<bool> has_norm_;
};

}  // namespace kernel

#endif  // KERNEL_LIB_CACHE_H__
