
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
// Author: rostami@google.com (Afshin Rostamizadeh)

#ifndef KERNEL_LIB_LIBLINEAR_H_
#define KERNEL_LIB_LIBLINEAR_H_

#include "kernel/lib/dense-matrix.h"
#include "kernel/lib/semidense-matrix.h"
#include "kernel/lib/dense-vector.h"
#include "kernel/lib/sparse-vector.h"

// TODO(allauzen): remove this include and add relevant forward
// declarations instead.
#include "linear.h"


// A wrapper to call the liblinear train command, using OpenKernel
// data structures.
namespace kernel {

// Modified versions of functions and structures taken from liblinear.
void solve_linear_c_svc(const problem *prob, double *w, double *alpha,
                        double eps, double Cp, double Cn,
                        double *f, // linear term coefficients
                        int solver_type);

// OpenKernel wrapper function.
// Solves the l1-loss svm linear svm problem with no additional offset
// feature.
// INPUT:
// data - a Dense or SemiDense matrix with instances as rows.
// labels - a DenseVector of labels for each instance,
// reg_param - the SVM regularization parameter "C", which upper bounds
//   the size of each dual variable.
// OUTPUT:
// dual_vars - hold the optimal dual variables on output.
template <class M>
void LibLinearTrain(const M &data,
                    const DenseVector<typename M::ElementType> &labels,
                    const double &reg_param,
                    DenseVector<typename M::ElementType> *dual_vars) {
  DenseVector<typename M::ElementType> f;
  f.Resize(labels.Size(), -1);
  LibLinearTrain(data, labels, reg_param, f, dual_vars);
}


// Same as previous function, but allows for arbitrary linear
// coefficients in the QP problem.
template <class M>
void LibLinearTrain(const M &data,
                    const DenseVector<typename M::ElementType> &labels,
                    const double &reg_param,
                    const DenseVector<typename M::ElementType> &f,
                    DenseVector<typename M::ElementType> *dual_vars) {
  CHECK(labels.Size() == data.NumRow());
  struct problem *prob = new struct problem;
  prob->l = data.NumRow();
  prob->n = data.NumColumn();
  prob->bias = -1;

  prob->y = new int[labels.Size()];
  for (size_t i = 0; i < labels.Size(); ++i)
    if (labels.Element(i) > 0)
      prob->y[i] = 1;
    else
      prob->y[i] = -1;

  prob->x = new struct feature_node*[data.NumRow()];
  DataToLibLinear(data, prob->x);

  double *w;
  w = new double[prob->n];
  double *alpha;
  alpha = new double[prob->l];
  double *f_array;
  f_array = new double[prob->l];
  for (size_t i = 0; i < prob->l; ++i)
    f_array[i] = f.Element(i);
  //solve_linear_c_svc(prob, w, alpha, eps, Cp, Cn, L1LOSS_SVM_DUAL);
  solve_linear_c_svc(prob, w, alpha, 1e-4, reg_param, reg_param, f_array,
                     L1LOSS_SVM_DUAL);

  dual_vars->Resize(data.NumRow(), 0.0);
  for (size_t i = 0; i < dual_vars->Size(); ++i)
    dual_vars->SetElement(i, alpha[i]);

  for (size_t i = 0; i < prob->l; ++i)
    delete[] prob->x[i];
  delete[] prob->x;
  delete[] prob->y;
  delete prob;
  delete[] w;
  delete[] alpha;
}

// Convert an OpenKernel DenseMatrix to an array of liblinear feature_nodes.
template <class T> inline
void DataToLibLinear(const DenseMatrix<T> &data, feature_node **x) {
  for (size_t i = 0; i < data.NumRow(); ++i) {
    x[i] = new struct feature_node[data.NumColumn() + 1];
    for (size_t j = 0; j < data.NumColumn(); ++j) {
      x[i][j].index = j + 1;
      x[i][j].value = data.Element(i,j);
    }
    x[i][data.NumColumn()].index = -1;
  }
}

// Convert an OpenKernel SparseMatrix (i.e. SemiDenseMatrix<SparseVector>)
// to an array of liblinear feature_nodes.
template <class T> inline
void DataToLibLinear(
    const SemiDenseMatrix<SparseVector<T>, DENSE_BY_ROW> &data,
    feature_node **x) {
  SparseVector<T> row;
  for (size_t i = 0; i < data.NumRow(); ++i) {
    row = data.Row(i);
    x[i] = new struct feature_node[row.Size() + 1];
    for (size_t j = 0; j < row.Size(); ++j) {
      x[i][j].index = row.Pair(j).first + 1;
      x[i][j].value = row.Pair(j).second;
    }
    x[i][row.Size()].index = -1;
  }
}

} // namespace kernel

#endif  // KERNEL_LIB_LIBLINEAR_H_
