// sparse-matrix.h
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
// Author: allauzen@google.com (Cyril Allauzen)
//
// \file
// A simple reference-counted sparse matrix representation.

#ifndef KERNEL_LIB_SPARSE_MATRIX_H__
#define KERNEL_LIB_SPARSE_MATRIX_H__

#include <vector>

#include "kernel/lib/dense-matrix.h"
#include "kernel/lib/sparse-vector.h"
#include "kernel/lib/kernel.h" // change to min req

namespace kernel {

enum MajorType { ROW_MAJOR = 0, COLUMN_MAJOR = 1 };

#define TRANSPOSE_MAJOR_TYPE(A) \
  ((A) == ROW_MAJOR ? COLUMN_MAJOR : ROW_MAJOR)

//
// SPARSE MATRIX CLASS
//

// Forward declaration
template <MajorType A, class T, class I = uint32>
class SparseMatrix;
template <class T, class I>
class SparseMatrixData;
template <class T, class I>
class SparseMatrixSplice;

// SparseMatrixBase:
template <MajorType A, class T, class I>
class SparseMatrixBase {
 public:
  typedef T ElementType;
  typedef I IndexType;
  typedef SparseMatrixBase<TRANSPOSE_MAJOR_TYPE(A), T, I> TransposeMatrix;
  typedef SparseMatrixSplice<T, I> VectorType;

  friend class SparseMatrixBase<TRANSPOSE_MAJOR_TYPE(A), T, I>;

  SparseMatrixBase() : data_(new SparseMatrixData<T, I>) {}

  SparseMatrixBase(const SparseMatrixBase<A, T, I> &base) : data_(base.data_) {
    data_->IncrRefCount();
  }

  virtual ~SparseMatrixBase() {
    if (!data_->DecrRefCount())
      delete data_;
  }

  SparseMatrixBase<A, T, I> &operator=(const SparseMatrixBase<A, T, I> &m) {
    if (this != &m) {
      if (!data_->DecrRefCount()) delete data_;
      m.data_->IncrRefCount();
      data_ = m.data_;
    }
    return *this;
  }

  void Transpose(TransposeMatrix *m) const {
    *m = TransposeMatrix(data_);
  }

  VectorType Vector(size_t i) const {
    return SparseMatrixSplice<T, I>(data_, data_->Index(i),
                                    data_->Index(i + 1));
  }
  //VectorType Vector(size_t i) const { return data_->Vector(i); }

  size_t Size() const { return data_->Size(); }

  size_t NumNonZeros() const { return data_->NumNonZeros(); }

  size_t MaxSparseIndex() const { return data_->MaxSparseIndex(); }

  template <class V>
  void AddVector(const V &vec) { data_->AddVector(vec); }

  void Clear() { data_->Clear(); }

  void Reserve(size_t num_nonzeros, size_t num_vectors) {
    data_->Reserve(num_nonzeros, num_vectors);
  }

 protected:
  SparseMatrixBase(SparseMatrixData<T, I> *data) : data_(data) {
    data_->IncrRefCount();
  }

  void MutateCheck() {
    // Copy on write
    if (data_->RefCount() > 1) {
      data_->DecrRefCount();
      data_ = new SparseMatrixData<T, I>(*data_);
    }
  }

 private:
  SparseMatrixData<T, I> *data_;
};


template <class T, class I>
class SparseMatrix<ROW_MAJOR, T, I>
    : public SparseMatrixBase<ROW_MAJOR, T, I> {
 public:
  typedef T ElementType;
  typedef I IndexType;
  typedef SparseMatrix<COLUMN_MAJOR, T, I> TransposeMatrix;
  typedef SparseMatrixSplice<T, I> RowType;

  using SparseMatrixBase<ROW_MAJOR, T, I>::Transpose;
  using SparseMatrixBase<ROW_MAJOR, T, I>::operator=;
  using SparseMatrixBase<ROW_MAJOR, T, I>::Size;
  using SparseMatrixBase<ROW_MAJOR, T, I>::MaxSparseIndex;
  using SparseMatrixBase<ROW_MAJOR, T, I>::Vector;

  SparseMatrix() {}
  SparseMatrix(const SparseMatrixBase<ROW_MAJOR, T, I> &base)
      : SparseMatrixBase<ROW_MAJOR, T, I>(base) {}

  size_t NumRow() const { return Size(); }
  size_t NumColumn() const { return MaxSparseIndex() + 1; }
  RowType Row(size_t i) const { return Vector(i); }
  template <class V> void AddRow(const V &row) { AddVector(row); }
};

template <class T, class I>
class SparseMatrix<COLUMN_MAJOR, T, I>
    : public SparseMatrixBase<COLUMN_MAJOR, T, I> {
 public:
  typedef T ElementType;
  typedef I IndexType;
  typedef SparseMatrix<ROW_MAJOR, T, I> TransposeMatrix;
  typedef SparseMatrixSplice<T, I> ColumnType;

  using SparseMatrixBase<COLUMN_MAJOR, T, I>::Transpose;
  using SparseMatrixBase<COLUMN_MAJOR, T, I>::operator=;
  using SparseMatrixBase<COLUMN_MAJOR, T, I>::Size;
  using SparseMatrixBase<COLUMN_MAJOR, T, I>::MaxSparseIndex;
  using SparseMatrixBase<COLUMN_MAJOR, T, I>::Vector;

  SparseMatrix() {};
  SparseMatrix(const SparseMatrixBase<COLUMN_MAJOR, T, I> &base)
      : SparseMatrixBase<COLUMN_MAJOR, T, I>(base) {}

  size_t NumColumn() const { return Size(); }
  size_t NumRow() const { return MaxSparseIndex() + 1; }
  ColumnType Column(size_t i) const { return Vector(i); }
  template <class V> void AddColumn(const V &col) { AddVector(col); }
};



template <class T, class I>
class SparseMatrixData {
 public:
  typedef SparseMatrixSplice<T, I> VectorType;

  SparseMatrixData() : max_sparse_index_(0) {
    indexes_.push_back(0);
  }

  explicit SparseMatrixData(const SparseMatrixData<T, I> &data)
      : pairs_(data.pairs_), indexes_(data.indexes_),
        max_sparse_index_(data.max_sparse_index_) {}

  ~SparseMatrixData() {}

  size_t Size() const { return indexes_.size() - 1; }

  size_t MaxSparseIndex() const { return max_sparse_index_; }

  size_t NumNonZeros() const { return pairs_.size(); }

  void Clear() {
    pairs_.clear();
    indexes_.resize(1);
  }

  void Reserve(size_t num_nonzeros, size_t num_vectors) {
    pairs_.reserve(num_nonzeros);
    indexes_.reserve(num_vectors + 1);
  }

  I Index(size_t i) const { return indexes_[i]; }

  const pair<I, T> &Pair(size_t i) const { return pairs_[i]; }

  template <class V>
  void AddVector(const V &v) {
    for (size_t i = 0; i < v.Size();  ++i)
      pairs_.push_back(v.Pair(i));
    indexes_.push_back(pairs_.size());
    if (v.MaxIndex() > max_sparse_index_)
      max_sparse_index_ = v.MaxIndex();
  }

  // TODO: Reserve()

  int RefCount() const { return ref_count_.count(); }
  int IncrRefCount() { return ref_count_.Incr(); }
  int DecrRefCount() { return ref_count_.Decr(); }

 private:
  vector<pair<I, T> > pairs_;
  vector<I> indexes_;
  size_t max_sparse_index_;
  RefCounter ref_count_;

  void operator=(const SparseMatrixData<T, I> &);  // disallow
};



template <class T, class I>
class SparseMatrixSplice {
 public:
  typedef T ElementType;
  typedef I IndexType;

  SparseMatrixSplice(SparseMatrixData<T, I> *data,
                     size_t begin,
                     size_t end)
      : data_(data), begin_(begin), end_(end) {
    data_->IncrRefCount();
  }

  SparseMatrixSplice(const SparseMatrixSplice<T, I> &splice)
      : data_(splice.data_), begin_(splice.begin_), end_(splice.end_) {
    data_->IncrRefCount();
  }

  ~SparseMatrixSplice() {
    if (!data_->DecrRefCount())
      delete data_;
  }

  size_t Size() const { return end_ - begin_; }

  const pair<I, T> &Pair(size_t i) const {
    return data_->Pair(begin_ + i);
  }

 private:
  SparseMatrixData<T, I> *data_;
  size_t begin_;
  size_t end_;
};

//
// MATRIX OPERATIONS
//

// Equality tests
template <MajorType A, class T, class I> inline
bool Equal(const SparseMatrix<A, T, I> &a,
           const SparseMatrix<A, T, I> &b) {
  if (a.Size() != b.Size())
    return false;

  bool flag = true;
  for (size_t i = 0; i < a.Size(); ++i)
    if (!Equal(a.Vector(i), b.Vector(i)))
      flag = false;

  return flag;
}

template <MajorType A, typename T, typename I> inline
bool ApproxEqual(const SparseMatrix<A, T, I> &a,
                 const SparseMatrix<A, T, I> &b,
                 T delta) {
  if (a.Size() != b.Size())
    return false;

  bool flag = true;
  for (size_t i = 0; i < a.Size(); ++i)
    if (!ApproxEqual(a.Vector(i), b.Vector(i), delta))
      flag = false;
  return flag;
}

// Construct a sparse diagonal matrix, with diagonal equal to
// diag_vector;
template <MajorType A, typename T, typename I> inline
void Diagonal(const DenseVector<T> &diag_vector,
              SparseMatrixBase<A, T, I> *m) {
  for (size_t i = 0; i < diag_vector.Size(); ++i) {
    SparseVector<T, I> temp;
    temp.AddElement(i, diag_vector.Element(i));
    m->AddVector(temp);
  }
}

// C = A * B
// with A dense by row and B dense by column
template <typename T, typename I> inline
void Product(const SparseMatrix<ROW_MAJOR, T, I> &a,
             const SparseMatrix<COLUMN_MAJOR, T, I> &b,
             DenseMatrix<T> *c) {
  c->Reset(a.NumRow(), b.NumColumn(), 0.0F);
  for (size_t i = 0; i < a.NumRow(); ++i)
    for (size_t j = 0; j < b.NumColumn(); ++j )
      c->SetElement(i, j, DotProduct(a.Row(i), b.Column(j)));
}

// C = A * B
// with A dense by row and B dense by column
// return a semidense matrix
template <typename T, typename I> inline
void Product(const SparseMatrix<ROW_MAJOR, T, I> &a,
             const SparseMatrix<COLUMN_MAJOR, T, I> &b,
             SparseMatrix<ROW_MAJOR, T, I> *c) {
  c->Clear();
  for (size_t i = 0; i < a.NumRow(); ++i) {
    SparseVector<T, I> temp;
    for (size_t j = 0; j < b.NumColumn(); ++j ) {
      SparseMatrixSplice<T,I> row(a.Row(i));
      SparseMatrixSplice<T,I> column(b.Column(j));
      temp.AddElement(j, DotProduct(row, column));
      //      temp.AddElement(j, DotProduct<SparseMatrixSplice<T, I>,
      //                SparseMatrixSplice<T,I> >(a.Row(i), b.Column(j)));
    }
    c->AddRow(temp);
  }
}

// C = A * B
// with A dense by column and B dense by row
template <typename T, typename I> inline
void Product(const SparseMatrix<COLUMN_MAJOR, T, I> &a,
             const SparseMatrix<ROW_MAJOR, T, I> &b,
             DenseMatrix<T> *c) {

  CHECK(a.NumColumn() == b.NumRow());
  c->Reset(a.NumRow(), b.NumColumn(), 0);
  for (size_t k = 0; k < a.NumColumn() ; ++k) {
    const SparseMatrixSplice<T, I> &column = a.Column(k);
    const SparseMatrixSplice<T, I> &row = b.Row(k);
    for (size_t i = 0; i < column.Size(); ++i)
      for (size_t j = 0; j < row.Size(); ++j) {
        pair<size_t, T>
            pair_i = column.Pair(i), pair_j = row.Pair(j);
        c->SetElement(pair_i.first, pair_j.first,
                      c->Element(pair_i.first, pair_j.first) +
                      pair_i.second * pair_j.second);
      }
  }
}

//
// MATRIX-VECTOR OPERATIONS
//

// B = A * diag(v)
template <typename T, typename I> inline
void DiagonalProduct(const SparseMatrix<ROW_MAJOR, T, I> &A,
                     const DenseVector<T> &v,
                     SparseMatrix<ROW_MAJOR, T, I> *B) {
  B->Clear();
  for (size_t i = 0; i < A.NumRow(); ++i) {
    SparseVector<T> temp;
    PointwiseProduct(A.Row(i), v, &temp);
    B->AddRow(temp);
  }
}

// u = B * v
// B is dense by column.
template <typename T, typename I> inline
void Product(const SparseMatrix<COLUMN_MAJOR, T, I> &a,
             const DenseVector<T> &v,
             DenseVector<T> *u) {
  CHECK(v.Size() == a.NumColumn());
  u->Clear();
  if (u->Size() < a.NumRow())
    u->Resize(a.NumRow(), 0);

  pair<size_t, T> p;
  for (size_t j = 0; j < a.NumColumn(); ++j) {
    SparseMatrixSplice<T, I> col(a.Column(j));
    for (size_t i = 0; i < col.Size(); ++i) {
      p = col.Pair(i);
      u->SetElement(p.first, u->Element(p.first)
                    + v.Element(j) * p.second);
    }
  }
}

// u = B * v
// B is dense by row.
template <typename T, typename I> inline
void Product(const SparseMatrix<ROW_MAJOR, T, I> &a,
             const DenseVector<T> &v,
             DenseVector<T> *u) {
  if (u->Size() != a.NumRow())
    u->Resize(a.NumRow(), 0);

  SparseVector<T> row;
  pair<size_t, T> p;
  for (size_t i = 0; i < a.NumRow(); ++i)
    u->SetElement(i, DotProduct(a.Row(i), v));
}


}  // namespace kernel

#endif  // KERNEL_LIB_SPARSE_MATRIX_H__
